//-------------------------------------------------------------------------------------------------
// ETH200 - Modul
//
// Zum senden von Steuerbefehle vom RFM12 an das ETH200 von ELV oder eQ3 von Conrad
//
// basierend auf Grundlagen aus dem Thread http://www.mikrocontroller.net/topic/172034
//
// Autor       : Michael Walischewski
// erstellt am : 25.04.2011
//
// Das Übertragungsprotokoll ist Manchester codiert
//
// nach Aufruf der Funktion ETH_Message muss der RFM sofern er für andere Zwecke genutzt
// wird wieder initialisiert werden !
//
// Aufruf :
//          ETH_Message(0x40,0x04,0x108F); // command 040 , +2 Grad Temperaturänderung, Gerät 0x108F
//          EthSetTemp(15,0x108F);         // Temperatur von Thermostat 0x108F auf 15 x 0.5° Grad setzen
//          EthLearn(0x108F)               // Thermostat soll auf die Nummer 0x108F hören
//---------------------------------------------------------------------------------------------------

#include <avr/io.h>
#include <util/delay.h>
#include "../Utils/RFM12.h"

#define F_CPU 16000000UL


void ETH_RFM12Init() {

    writeCmd(0x80A5); // TX aktiv,868band,11.0pF
    writeCmd(0xA67C); // 868,3 MHz
    writeCmd(0xC623); // 9600 Baud !!
    writeCmd(0xC4A7); // Range limit +7fres to -8fres, Max DEviation +35/-40Khz, Offset VDI=h
    writeCmd(0xC238); // Datafilter auf analog filter
    writeCmd(0x9820); // volles ROHR und df(fsk) auf 15KHz
    writeCmd(0x8279); // Enable baseband, PLL, PA, TX, synthesizer, crystal oscillator, disable clock output
    writeCmd(0xE000); // Wake-up Timer off
    writeCmd(0xC800); // low Duty-Cycle off
    writeCmd(0xC000); // 1.66MHz,2.2V
    writeCmd(0x0000); // Status auslesen
}

// Diese CRC-Routine wurde von Dietmar Weisser aus dem Forum www.mikrocontroller.net übernommen
// Thread ETH200 CRC
//
uint16_t calcCRC16r( uint16_t c,uint16_t crc, uint16_t mask)  // reverser crc!!!!!!
{
  unsigned char i;
  for(i=0;i<8;i++)
  {
    if((crc ^ c) & 1) { crc=(crc>>1)^mask; }
    else crc>>=1;
    c>>=1;
  };
 return(crc);
}

unsigned char reverseByte(unsigned char Data)
{
  unsigned char i, gedreht=0;

  for(i=0;i<8;i++){
     if(Data&(1<<i)){
        gedreht = gedreht <<1 | 1;
     }
     else{
        gedreht = gedreht <<1;
     }
  }
  return gedreht;
}

unsigned char EthMessage(unsigned char cmd, unsigned char Wert,uint32_t Adresse)
{
    static unsigned char cnt;                      // Paket counter
	unsigned char ethPaket[30];                    // Fertiges Manchestercodierte ETH200 Paket
	unsigned char buf[10],revbuf[10];              // revbuf beinhaltet das Paket mit reverse angeordneten Bits
	unsigned char b,ethPaketi,bitcnt,Einscnt;
	uint16_t crc=0,i=0;
	int j;


	ethPaketi = 0;                 // Paketzähler
	Einscnt   = 0;                 // Zähler für die aufeinanderfolgenden einsen
	b         = 0;
	bitcnt    = 0;

    cnt++;
    if (cnt>255)cnt=1;



	buf[0] = 0x7e;                   // init
	buf[1] = cnt;                   // spielt keine Rolle wenn ein Bitstream zeitlich versetzt mit der gleichen PaketID gesendet wird.
	buf[2] = 0x10;                   // Paketlänge inkl CRC16
	buf[3] = (uint8_t)(Adresse>>16); //0x00; //(uint8_t)(Adresse>>16);  // Adresse
	buf[4] = (uint8_t)(Adresse>>8);  //0x12; //(uint8_t)(Adresse>>8);
	buf[5] = (uint8_t)(Adresse);     //0x8F; //(uint8_t)(Adresse);
	buf[6] = cmd;                    //0x40; //msg.cmd;                 // Commando 40 - Temperatur ändern !, 42 - Tagbetrieb, 43 - Nachtbetrieb
	buf[7] = Wert;                   //31 wäre zwar maximal von 5 -> 29.5 aber ich wollts mal wissen ;)     // Grad der Erhöhung, 01 -> 0.5 Grad, 02 -> 1 Grad, FF -> -0.5 , FE -> -1.0


    crc=0xc11f;
	for(i=0;i<8;i++){
		crc=calcCRC16r(buf[i],crc,0x8408);
	};

	buf[8]=(uint8_t)crc;
	buf[9]=(uint8_t)(crc>>8);


   // alle Bits pro Bytes reversen
   for(i=0;i<10;i++){
      revbuf[i]=reverseByte(buf[i]);
   }

    // Manchestercodierung
    // aus 1- mach 2Bytes. 1 => 10 , 0 => 01
    // bei 5 folgenden 1 was bei minus Temperaturen der FAll ist muss eine 0 eingeschoben werden und alle Bits verschieben sich ! Allerdings
    // nicht bei der Initialisierung 7E.

     // Alle 10 Paket durchlaufen !
    for (i=0;i<10;i++){
        // und alle 8 Bits vom Höherwertigen an
        for (j=7;j>-1;j--){
           //und nach der Hälfte haben wir ein Byte voll und machen mit dem nächsten weiter
           if (bitcnt==8){
              ethPaket[ethPaketi++] = b;
              b=0;
              bitcnt=0;
           }
          if((revbuf[i]&(1<<j)) && (Einscnt==4)){
               // 0 einschieben
               b=b<<1 | 0;
               b=b<<1 | 1;
               bitcnt +=2;
               Einscnt = 0;
               if (bitcnt==8){
                  ethPaket[ethPaketi++] = b;
                  b=0;
                  bitcnt=0;
               }
           }

           if(revbuf[i]&(1<<j)){
               b=b<<1 | 1;
               b=b<<1 | 0;
               if(i>0){
               Einscnt++;
               }
            }
            else{
               b=b<<1 | 0;
               b=b<<1 | 1;
               Einscnt = 0;
            }
            bitcnt = bitcnt + 2;
        }
     }
     if(bitcnt>0)ethPaket[ethPaketi++] = b;   // letztes angebrochene Byte sichern

     writeCmd(0x0000);
     writeCmd(0x8239);
     rfSend(0x55);        // Init damit sich der Tuner einschwingt
     rfSend(0x55);
     rfSend(0xAA);
     rfSend(0xAA);

     // Fummelwert, die 320 haben gut funktioniert beim ETH200, beim e-Q3 muss der Wert niedriger gewählt werden, Schätzungsweise auf ca. 290 !
     for(i=0;i<320;i++){
	    for(j=0;j<ethPaketi;j++){
	       rfSend(ethPaket[j]);
        }
     }

     rfSend(0x00);
     rfSend(0x00);
     _delay_ms(6);
     writeCmd(0x8219);
     writeCmd(0x0000);

   	 return 0;
}

//
// Da mit dem Protokoll der Fernbedienung nur eine relative Temperaturerhöhung um x Grad möglich ist
// und auch der Thermostat mit der Hand verdreht sein kann , setze ich die Temperatur erstmal komplett
// zurück , also 5 Grad und dann auf die gewünschte Temperatur

unsigned char EthSetTemp(unsigned char Wert,uint32_t Adresse){

   ETH_RFM12Init();
   _delay_ms(500);
   EthMessage(0x40,0xCA,Adresse);   // -30 Grad
   _delay_ms(1000);                 // etwas Zeit vergehen lassen, bevor jetzt gleich das nächste Paket versendet wird !
   EthMessage(0x40,Wert,Adresse);   // Offset von +5Grad und Wert in 0.5Schritten
   _delay_ms(100);

   // RFM12 wieder in den "normalen" Modus versetzen !
   rfInit();
   switch2RX();

   return 0;
}


unsigned char EthLearn(uint32_t Adresse){
   ETH_RFM12Init();
   _delay_ms(500);
   EthMessage(0x40,0x00,Adresse);

   rfInit();
   switch2RX();

   return 0;
}
